// File: Stepper.cpp
// Stepper control for A4983 or equivalent 
// Motor uses an A4983 stepperdriver with Enable, Dir and Step pins.
// Target: Arduino NANO
// 2015, jan@breem.nl.

#include "arduino.h"
#include "globals.h"

long ActualPosition; // actual position in motor-steps
long Traject, TrajectProgress, DecellerationTraject;
bool TrajectBusy, Accellerating, Decellerating, DirNegative;
int Speed, MaxSpeed, MinSpeed, Accell;
int Reload;
long EffectiveClock;

//********************************************************************************************************
void Motor_Step (bool Dir)
// The A4983 steps on a rising edge of the Step signal and requires a pulse of minimal 1 usec.
// I deliberately made the rising edge at the end of the routine, and the falling edge at the begin.
// The alternative, first high and immediately therafter low, may result in a too short pulse.
{
  PORTB &= ~0x20; // step low  
  if (Dir) 
  {
    PORTB |= 0x10; // DIR pin
    ActualPosition++;
  } 
  else 
  {
    PORTB &= ~0x10; // DIR pin
    ActualPosition--;
  } 
  PORTB |= 0x20; // step high
}

//********************************************************************************************************
void Motor_Enable (bool Yes)
{
if (Yes) PORTB &= ~0x08; // Low is enable
else PORTB |= 0x08;  // High is disable
}

//********************************************************************************************************
void EvalStepper(void) // Call by ISR(TIMER2_COMPA_vect) at 100 Hz 
{ 
  // Evaluates accelleration/decelleration and max speed
  unsigned int Divisor; 
  if (Accellerating) Speed += Accell;
  if (Decellerating) Speed -= Accell;
  if (Speed >= MaxSpeed) {Accellerating = false; Speed = MaxSpeed;}
  if (Speed <= MinSpeed) {Decellerating = false; Speed = MinSpeed;}
  Reload = EffectiveClock / Speed; 
} 

//********************************************************************************************************
ISR(TIMER2_COMPA_vect) // Called at 100 Hz
{
  EvalStepper();
}

//************************************************************************************************
void StartTraject (long NumberOfSteps, int StartSpd, int MaxSpd, int Accel)
{
  if (TrajectBusy) {Serial.println ("Busy"); return;} // already started 
  Serial.println ("Traject Started"); 
  TrajectBusy = true;     
  DirNegative = (NumberOfSteps < 0);
  Traject = abs(NumberOfSteps);
  TrajectProgress = 0;
  DecellerationTraject = 0;
  Accellerating = true;
  Decellerating = false;
  Motor_Enable (true);  
  Speed = StartSpd;  // StartSpeed in steps / sec;
  MinSpeed = 100;
  if (Speed < MinSpeed) Speed = MinSpeed;
  MaxSpeed = MaxSpd; 
  Accell = Accel;
  TIMSK1 = 0x02; // enable interrupts voor Timer 1 compare A   
}

//************************************************************************************************
void StopTraject (void)
{
  // Force decellerating, ending in a complete and nice stop
  TrajectProgress = Traject - DecellerationTraject;  
}

//************************************************************************************************
ISR(TIMER1_COMPA_vect) 
// Stepper interrupt to do a trajectory
// Interrupt frequency varies according to accelleration etc.
{
  DIAGpin20H  // Pin 20 shows the Interrupt-In-Service time  
  if (TrajectBusy) // Safety measure. Interrupt should normally not happen when Traject is !Busy
  {
    if (TrajectProgress++ >= Traject)  // Traject completed
    {
      TIMSK1 = 0; // disable stepper interrupt
      TrajectBusy = false;
      Accellerating = false;
      Decellerating = false;
      Motor_Enable(false);  
      Serial.println ("Traject finished");      
      return; 
    }
    // we need to have a decelleration traject as long as the accelleration traject
    if (Accellerating) DecellerationTraject++; 
    if (TrajectProgress >= (Traject - DecellerationTraject)) // Now start decelleration.
    {
      Accellerating = false;
      Decellerating = true;
    }
    Motor_Step(DirNegative);
    OCR1A = Reload;  // and set a possibly new step rate
    // Changeing the steprate inside the interrupt routine guarantees that the timer correcttly works with the new value.
    // If changed outside the interrupt routine it may happen that the timer is already past the new match value, so 
    // it will run to its end at 0xFFFF, causing the motor to stumble.  
  }
  DIAGpin20L // Pin 20 shows the Interrupt-In-Service time  
}


//********************************************************************************************************
void InitStepper (void)
{
  // Configure 16 bit Timer 1 for variable Step rate  
  TCCR1A = 0x00;  // Waveform mode 4, no PWM outputs
  TCCR1B = 0x0A;  // Internal clock div 8 gives 2 MHz 
  EffectiveClock = 2000000; // 2 MHz / Reload-value gives the exact steprate in steps / sec.
 
  // Configure 8 bit Timer2 for 100 Hz interrupts
  TCCR2A = 0x02;  // WaveformMode 2, no PWM outputs
  TCCR2B = 0x07;  // Internal clock div 1024 for 2 MHz
  TIMSK2 = 0x02;  // Interrupt on OCRA match   
  OCR2A = 156;    // divide by 156 for 100 Hz interrupts
 
  DDRB = 0xFF;    // all outputs
  Motor_Enable (false);
  
  UseDIAGpin19
  UseDIAGpin20
  UseDIAGpin21
  
  sei();  // enable interrupts;
  Serial.println ("Stepper Initialized"); 
}


